@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2018-2023. All Rights Reserved.
+%% Copyright Ericsson AB 2018-2024. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -20,11 +20,13 @@
 %% Purpose: Type definitions and utilities for the SSA format.
 
 -module(beam_ssa).
+-moduledoc false.
 -export([add_anno/3,get_anno/2,get_anno/3,
          between/4,
          clobbers_xregs/1,def/2,def_unused/3,
          definitions/2,
          dominators/2,dominators_from_predecessors/2,common_dominators/3,
+         eval_instr/1,
          flatmapfold_instrs/4,
          fold_blocks/4,
          fold_instrs/4,
@@ -48,7 +50,7 @@
               b_ret/0,b_br/0,b_switch/0,terminator/0,
               b_var/0,b_literal/0,b_remote/0,b_local/0,
               value/0,argument/0,label/0,
-              var_name/0,var_base/0,literal_value/0,
+              var_name/0,literal_value/0,
               op/0,anno/0,block_map/0,dominator_map/0,
               rename_map/0,rename_proplist/0,usage_map/0,
               definition_map/0,predecessor_map/0]).
@@ -78,8 +80,7 @@
 -type argument()   :: value() | b_remote() | b_local() | phi_value().
 -type label()      :: non_neg_integer().
 
--type var_name()   :: var_base() | {var_base(),non_neg_integer()}.
--type var_base()   :: atom() | non_neg_integer().
+-type var_name()   :: atom() | non_neg_integer().
 
 -type literal_value() :: atom() | integer() | float() | list() |
                          nil() | tuple() | map() | binary() | fun().
@@ -105,30 +106,38 @@
 %% To avoid the collapsing, change the value of SET_LIMIT to 50 in the
 %% file erl_types.erl in the dialyzer application.
 
--type prim_op() :: 'bs_extract' | 'bs_get_tail' | 'bs_init_writable' |
+-type prim_op() :: 'bs_create_bin' |
+                   'bs_extract' | 'bs_ensure' | 'bs_get_tail' | 'bs_init_writable' |
                    'bs_match' | 'bs_start_match' | 'bs_test_tail' |
                    'build_stacktrace' |
                    'call' | 'catch_end' |
-                   'extract' |
+                   'executable_line' | 'extract' |
                    'get_hd' | 'get_map_element' | 'get_tl' | 'get_tuple_element' |
                    'has_map_field' |
                    'is_nonempty_list' | 'is_tagged_tuple' |
                    'kill_try_tag' |
                    'landingpad' |
-                   'make_fun' | 'match_fail' | 'new_try_tag' | 'old_make_fun' |
+                   'make_fun' | 'match_fail' | 'new_try_tag' |
+                   'nif_start' |
                    'peek_message' | 'phi' | 'put_list' | 'put_map' | 'put_tuple' |
-                   'raw_raise' | 'recv_next' | 'remove_message' | 'resume' |
+                   'raw_raise' |
+                   'recv_marker_bind' |
+                   'recv_marker_clear' |
+                   'recv_marker_reserve' |
+                   'recv_next' | 'remove_message' | 'resume' |
+                   'update_tuple' | 'update_record' |
                    'wait_timeout'.
 
 -type float_op() :: 'checkerror' | 'clearerror' | 'convert' | 'get' | 'put' |
                     '+' | '-' | '*' | '/'.
 
 %% Primops only used internally during code generation.
--type cg_prim_op() :: 'bs_get' | 'bs_get_position' | 'bs_match_string' |
+-type cg_prim_op() :: 'bs_checked_get' | 'bs_checked_skip' |
+                      'bs_get' | 'bs_get_position' | 'bs_match_string' |
                       'bs_restore' | 'bs_save' | 'bs_set_position' | 'bs_skip' |
                       'copy' | 'match_fail' | 'put_tuple_arity' |
-                      'put_tuple_element' | 'put_tuple_elements' |
-                      'set_tuple_element' | 'succeeded'.
+                      'set_tuple_element' | 'succeeded' |
+                      'update_record'.
 
 -import(lists, [foldl/3,mapfoldl/3,member/2,reverse/1,sort/1]).
 
@@ -179,7 +188,6 @@ clobbers_xregs(#b_set{op=Op}) ->
         build_stacktrace -> true;
         call -> true;
         landingpad -> true;
-        old_make_fun -> true;
         peek_message -> true;
         raw_raise -> true;
         wait_timeout -> true;
@@ -221,6 +229,8 @@ no_side_effect(#b_set{op=Op}) ->
         put_tuple -> true;
         raw_raise -> true;
         {succeeded,guard} -> true;
+        update_record -> true;
+        update_tuple -> true;
         _ -> false
     end.
 
@@ -239,7 +249,7 @@ no_side_effect(#b_set{op=Op}) ->
     Count :: label(),
     Result :: {block_map(), label()}.
 
-insert_on_edges(Insertions, Blocks, Count) ->
+insert_on_edges(Insertions, Blocks, Count) when is_map(Blocks) ->
     %% Sort insertions to simplify the handling of duplicates.
     insert_on_edges_1(sort(Insertions), Blocks, Count).
 
@@ -329,7 +339,7 @@ is_loop_header(#b_set{op=Op}) ->
       Result :: predecessor_map().
 
 predecessors(Blocks) ->
-    P0 = [{S,L} || {L,Blk} <- maps:to_list(Blocks),
+    P0 = [{S,L} || L := Blk <- Blocks,
                    S <- successors(Blk)],
     P1 = sofs:relation(P0),
     P2 = sofs:rel2fam(P1),
@@ -381,7 +391,7 @@ normalize(#b_set{anno=Anno0,op={bif,Bif},args=Args}=Set) ->
                                #{1 := Type} ->
                                    Anno0#{arg_types => #{0 => Type}};
                                #{} ->
-                                   Anno0
+                                   Anno0#{arg_types => #{}}
                            end;
                        #{} ->
                            Anno0
@@ -435,20 +445,20 @@ successors(L, Blocks) ->
 -spec def(Ls, Blocks) -> Def when
       Ls :: [label()],
       Blocks :: block_map(),
-      Def :: ordsets:ordset(var_name()).
+      Def :: ordsets:ordset(b_var()).
 
-def(Ls, Blocks) ->
+def(Ls, Blocks) when is_map(Blocks) ->
     Blks = [map_get(L, Blocks) || L <- Ls],
     def_1(Blks, []).
 
 -spec def_unused(Ls, Used, Blocks) -> {Def,Unused} when
       Ls :: [label()],
-      Used :: ordsets:ordset(var_name()),
+      Used :: ordsets:ordset(b_var()),
       Blocks :: block_map(),
-      Def :: ordsets:ordset(var_name()),
-      Unused :: ordsets:ordset(var_name()).
+      Def :: ordsets:ordset(b_var()),
+      Unused :: ordsets:ordset(b_var()).
 
-def_unused(Ls, Unused, Blocks) ->
+def_unused(Ls, Unused, Blocks) when is_map(Blocks) ->
     Blks = [map_get(L, Blocks) || L <- Ls],
     Preds = sets:from_list(Ls, [{version, 2}]),
     def_unused_1(Blks, Preds, [], Unused).
@@ -470,7 +480,7 @@ def_unused(Ls, Unused, Blocks) ->
       Labels :: [label()],
       Blocks :: block_map(),
       Result :: {dominator_map(), numbering_map()}.
-dominators(Labels, Blocks) ->
+dominators(Labels, Blocks) when is_map(Blocks) ->
     Preds = predecessors(Blocks),
     dominators_from_predecessors(Labels, Preds).
 
@@ -478,7 +488,7 @@ dominators(Labels, Blocks) ->
       Labels :: [label()],
       Preds :: predecessor_map(),
       Result :: {dominator_map(), numbering_map()}.
-dominators_from_predecessors(Top0, Preds) ->
+dominators_from_predecessors(Top0, Preds) when is_map(Preds) ->
     Df = maps:from_list(number(Top0, 0)),
     [{0,[]}|Top] = [{L,map_get(L, Preds)} || L <- Top0],
 
@@ -492,17 +502,57 @@ dominators_from_predecessors(Top0, Preds) ->
 %%  and Dominators and Numbering as returned from dominators/1.
 
 -spec common_dominators([label()], dominator_map(), numbering_map()) -> [label()].
-common_dominators(Ls, Dom, Numbering) ->
+common_dominators(Ls, Dom, Numbering) when is_map(Dom) ->
     Doms = [map_get(L, Dom) || L <- Ls],
     dom_intersection(Doms, Numbering).
 
+
+%% eval_instr(Instr) -> #b_literal{} | any | failed.
+%%  Attempt to evaluate a BIF instruction. Returns a `#b_literal{}`
+%%  record if evaluation succeeded, `failed` if an exception was
+%%  raised, or `any` if the arguments were not literals or the BIF is
+%%  not pure.
+
+-spec eval_instr(b_set()) -> b_literal() | 'any' | 'failed'.
+
+eval_instr(#b_set{op={bif,Bif},args=Args}) ->
+    LitArgs = case Args of
+                  [#b_literal{val=Arg1}] ->
+                      [Arg1];
+                  [#b_literal{val=Arg1},#b_literal{val=Arg2}] ->
+                      [Arg1,Arg2];
+                  [#b_literal{val=Arg1},#b_literal{val=Arg2},#b_literal{val=Arg3}] ->
+                      [Arg1,Arg2,Arg3];
+                  _ ->
+                      none
+              end,
+    case LitArgs of
+        none ->
+            any;
+        _ ->
+            Arity = length(LitArgs),
+            case erl_bifs:is_pure(erlang, Bif, Arity) of
+                true ->
+                    try apply(erlang, Bif, LitArgs) of
+                        Result ->
+                            #b_literal{val=Result}
+                    catch error:_->
+                            failed
+                    end;
+                false ->
+                    any
+            end
+    end;
+eval_instr(_) ->
+    any.
+
 -spec fold_instrs(Fun, Labels, Acc0, Blocks) -> any() when
       Fun :: fun((b_set()|terminator(), any()) -> any()),
       Labels :: [label()],
       Acc0 :: any(),
       Blocks :: block_map().
 
-fold_instrs(Fun, Labels, Acc0, Blocks) ->
+fold_instrs(Fun, Labels, Acc0, Blocks) when is_map(Blocks) ->
     fold_instrs_1(Labels, Fun, Blocks, Acc0).
 
 %% mapfold_blocks(Fun, [Label], Acc, BlockMap) -> {BlockMap,Acc}.
@@ -515,7 +565,7 @@ fold_instrs(Fun, Labels, Acc0, Blocks) ->
       Acc :: any(),
       Blocks :: block_map(),
       Result :: {block_map(), any()}.
-mapfold_blocks(Fun, Labels, Acc, Blocks) ->
+mapfold_blocks(Fun, Labels, Acc, Blocks) when is_map(Blocks) ->
     foldl(fun(Lbl, A) ->
                   mapfold_blocks_1(Fun, Lbl, A)
           end, {Blocks, Acc}, Labels).
@@ -534,7 +584,7 @@ mapfold_blocks_1(Fun, Lbl, {Blocks0, Acc0}) ->
       Blocks0 :: block_map(),
       Blocks :: block_map().
 
-mapfold_instrs(Fun, Labels, Acc0, Blocks) ->
+mapfold_instrs(Fun, Labels, Acc0, Blocks) when is_map(Blocks) ->
     mapfold_instrs_1(Labels, Fun, Blocks, Acc0).
 
 -spec flatmapfold_instrs(Fun, Labels, Acc0, Blocks0) -> {Blocks,Acc} when
@@ -545,7 +595,7 @@ mapfold_instrs(Fun, Labels, Acc0, Blocks) ->
       Blocks0 :: block_map(),
       Blocks :: block_map().
 
-flatmapfold_instrs(Fun, Labels, Acc0, Blocks) ->
+flatmapfold_instrs(Fun, Labels, Acc0, Blocks) when is_map(Blocks) ->
     flatmapfold_instrs_1(Labels, Fun, Blocks, Acc0).
 
 -type fold_fun() :: fun((label(), b_blk(), any()) -> any()).
@@ -560,7 +610,7 @@ flatmapfold_instrs(Fun, Labels, Acc0, Blocks) ->
       Acc0 :: any(),
       Blocks :: #{label():=b_blk()}.
 
-fold_blocks(Fun, Labels, Acc0, Blocks) ->
+fold_blocks(Fun, Labels, Acc0, Blocks) when is_map(Blocks) ->
     fold_blocks_1(Labels, Fun, Blocks, Acc0).
 
 %% linearize(Blocks) -> [{BlockLabel,#b_blk{}}].
@@ -574,7 +624,7 @@ fold_blocks(Fun, Labels, Acc0, Blocks) ->
       Blocks :: block_map(),
       Linear :: [{label(),b_blk()}].
 
-linearize(Blocks) ->
+linearize(Blocks) when is_map(Blocks) ->
     Seen = sets:new([{version, 2}]),
     {Linear0,_} = linearize_1([0], Blocks, Seen, []),
     Linear = fix_phis(Linear0, #{}),
@@ -592,7 +642,7 @@ rpo(Blocks) ->
       Blocks :: block_map(),
       Labels :: [label()].
 
-rpo(From, Blocks) ->
+rpo(From, Blocks) when is_map(Blocks) ->
     Seen = sets:new([{version, 2}]),
     {Ls,_} = rpo_1(From, Blocks, Seen, []),
     Ls.
@@ -609,7 +659,7 @@ rpo(From, Blocks) ->
       Blocks :: block_map(),
       Labels :: [label()].
 
-between(From, To, Preds, Blocks) ->
+between(From, To, Preds, Blocks) when is_map(Preds), is_map(Blocks) ->
     %% Gather the predecessors of `To` and then walk forward from `From`,
     %% skipping the blocks that don't precede `To`.
     %%
@@ -627,7 +677,7 @@ between(From, To, Preds, Blocks) ->
 
 rename_vars(Rename, Labels, Blocks) when is_list(Rename) ->
     rename_vars(maps:from_list(Rename), Labels, Blocks);
-rename_vars(Rename, Labels, Blocks) when is_map(Rename)->
+rename_vars(Rename, Labels, Blocks) when is_map(Rename), is_map(Blocks) ->
     Preds = sets:from_list(Labels, [{version, 2}]),
     F = fun(#b_set{op=phi,args=Args0}=Set) ->
                 Args = rename_phi_vars(Args0, Preds, Rename),
@@ -658,7 +708,7 @@ rename_vars(Rename, Labels, Blocks) when is_map(Rename)->
       Blocks :: block_map(),
       Count :: label().
 
-split_blocks(Ls, P, Blocks, Count) ->
+split_blocks(Ls, P, Blocks, Count) when is_map(Blocks) ->
     split_blocks_1(Ls, P, Blocks, Count).
 
 -spec trim_unreachable(SSA0) -> SSA when
@@ -676,7 +726,7 @@ trim_unreachable(Blocks) when is_map(Blocks) ->
 trim_unreachable([_|_]=Blocks) ->
     trim_unreachable_1(Blocks, sets:from_list([0], [{version, 2}])).
 
--spec used(b_blk() | b_set() | terminator()) -> [var_name()].
+-spec used(b_blk() | b_set() | terminator()) -> [b_var()].
 
 used(#b_blk{is=Is,last=Last}) ->
     used_1([Last|Is], ordsets:new());
@@ -707,7 +757,7 @@ definitions(Labels, Blocks) ->
 -spec uses(Labels, Blocks) -> usage_map() when
       Labels :: [label()],
       Blocks :: block_map().
-uses(Labels, Blocks) ->
+uses(Labels, Blocks) when is_map(Blocks) ->
     fold_blocks(fun fold_uses_block/3, Labels, #{}, Blocks).
 
 fold_uses_block(Lbl, #b_blk{is=Is,last=Last}, UseMap0) ->
